package me.hammerle.code;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import me.hammerle.exceptions.IllegalStringException;
import java.time.ZonedDateTime;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.List;
import java.awt.Font;
import javax.imageio.ImageIO;
import me.hammerle.exceptions.IllegalColorException;
import me.hammerle.graphics.*;
import me.hammerle.snuviscript.SnuviScript;

public class CodeParser 
{    
    @SuppressWarnings("unchecked")
    public static Object parseFunction(Script sd, String function, Object[] args, int jump)
    {
        try
        {
            switch(function.toLowerCase())
            {       
                // Rechenoperationen
                /*case "add":
                    return args.stream().mapToDouble(s -> ScriptUtils.getDouble(s)).sum();
                case "sub":
                    return ScriptUtils.getDouble(args.get(0)) - ScriptUtils.getDouble(args.get(1));
                case "mul":
                    return ScriptUtils.getDouble(args.get(0)) * ScriptUtils.getDouble(args.get(1));
                case "div":
                    return ScriptUtils.getDouble(args.get(0)) / ScriptUtils.getDouble(args.get(1));
                case "math.mod":
                    return ScriptUtils.getDouble(args.get(0)) % ScriptUtils.getDouble(args.get(1));
                case "math.abs":
                    return Math.abs(ScriptUtils.getDouble(args.get(0)));
                case "math.pow":
                    return Math.pow(ScriptUtils.getDouble(args.get(0)), ScriptUtils.getDouble(args.get(1)));
                case "math.root":
                    return Math.pow(ScriptUtils.getDouble(args.get(0)), 1d / ScriptUtils.getDouble(args.get(1)));
                case "math.sin":
                    return Math.sin(ScriptUtils.getDouble(args.get(0)));
                case "math.cos":
                    return Math.cos(ScriptUtils.getDouble(args.get(0)));
                case "math.e":
                    return Math.E;
                case "math.pi":
                    return Math.PI;
                case "math.ln":
                    return Math.log(ScriptUtils.getDouble(args.get(0)));
                case "math.log":
                    return Math.log10(ScriptUtils.getDouble(args.get(0)));
                case "math.random":
                    return ScriptUtils.randomInt(ScriptUtils.getInteger(args.get(0)), ScriptUtils.getInteger(args.get(1)));
                case "math.round":
                    return ((Long) Math.round(ScriptUtils.getDouble(args.get(0)))).intValue();
                case "math.rounddown":
                    return Math.floor(ScriptUtils.getDouble(args.get(0)));
                case "math.roundup":
                    return Math.ceil(ScriptUtils.getDouble(args.get(0)));
                case "math.roundcomma":
                    return new BigDecimal(ScriptUtils.getDouble(args.get(0))).setScale(ScriptUtils.getInteger(args.get(1)), RoundingMode.HALF_UP);

                // Variablen
                case "getvar":
                    return sd.getVar(args.get(0));
                case "setvar":
                    sd.setVar(args.get(0), args.get(1)); break;
                case "removevar":
                    sd.removeVar(args.get(0)); break;          

                // Listen - Befehle                         
                case "list.new":                                               
                    sd.setVar(args.get(0), new ArrayList<>()); break; 
                case "list.exists":                                               
                    return sd.getVar(args.get(0)) instanceof List;
                case "list.add":  
                    ((List) sd.getVar(args.get(0))).add(args.get(1)); break;
                case "list.remove":          
                    ((List) sd.getVar(args.get(0))).remove(args.get(1)); break;
                case "list.removeindex":           
                    ((List) sd.getVar(args.get(0))).remove((int) args.get(1)); break;                
                case "list.contains":                           
                    return ((List) sd.getVar(args.get(0))).contains(args.get(1));
                case "list.size":                           
                    return ((List) sd.getVar(args.get(0))).size();
                case "list.get":                           
                    return ((List) sd.getVar(args.get(0))).get((int) args.get(1));
                case "list.set":                           
                    ((List) sd.getVar(args.get(0))).set((int) args.get(1), args.get(2)); break;
                case "list.indexof":                           
                    return ((List) sd.getVar(args.get(0))).indexOf(args.get(1));
                case "list.sort": 
                    sortList(args, sd); break;
                case "list.reverse":          
                    Collections.reverse((List<Object>) sd.getVar(args.get(0))); break;
                case "list.shuffle":                           
                    Collections.shuffle((List<Object>) sd.getVar(args.get(0))); break;

                // Maps - Befehle                         
                case "map.new":                                               
                    sd.setVar(args.get(0), new HashMap<>()); break;
                case "map.exists":                                               
                    return sd.getVar(args.get(0)) instanceof HashMap; 
                case "map.add":                           
                    ((HashMap) sd.getVar(args.get(0))).put(args.get(1), args.get(2)); break;
                case "map.remove":                           
                    ((HashMap) sd.getVar(args.get(0))).remove(args.get(1)); break;
                case "map.contains":                         
                    return ((HashMap) sd.getVar(args.get(0))).containsKey(args.get(1));
                case "map.get":                           
                    return ((HashMap) sd.getVar(args.get(0))).get(args.get(1));

                // Grafik
                case "repaint":
                    SnuviScript.screen.graphics.repaint(); break;
                case "drawline":
                    SnuviScript.screen.graphics.addGraphicObject(ScriptUtils.getInteger(args.get(0)), 
                            new Line(ScriptUtils.getColor(args.get(1)),
                            ScriptUtils.getInteger(args.get(2)), ScriptUtils.getInteger(args.get(3)),
                            ScriptUtils.getInteger(args.get(4)), ScriptUtils.getInteger(args.get(5)))); break;
                case "drawoval":
                    SnuviScript.screen.graphics.addGraphicObject(ScriptUtils.getInteger(args.get(0)), 
                            new Oval(ScriptUtils.getColor(args.get(1)),
                            ScriptUtils.getInteger(args.get(2)), ScriptUtils.getInteger(args.get(3)),
                            ScriptUtils.getInteger(args.get(4)), ScriptUtils.getInteger(args.get(5)))); break;
                case "drawrec":
                    SnuviScript.screen.graphics.addGraphicObject(ScriptUtils.getInteger(args.get(0)), 
                            new Rectangle(ScriptUtils.getColor(args.get(1)),
                            ScriptUtils.getInteger(args.get(2)), ScriptUtils.getInteger(args.get(3)),
                            ScriptUtils.getInteger(args.get(4)), ScriptUtils.getInteger(args.get(5)))); break;
                case "drawfulloval":
                    SnuviScript.screen.graphics.addGraphicObject(ScriptUtils.getInteger(args.get(0)), 
                            new FullOval(ScriptUtils.getColor(args.get(1)),
                            ScriptUtils.getInteger(args.get(2)), ScriptUtils.getInteger(args.get(3)),
                            ScriptUtils.getInteger(args.get(4)), ScriptUtils.getInteger(args.get(5)))); break;
                case "drawfullrec":
                    SnuviScript.screen.graphics.addGraphicObject(ScriptUtils.getInteger(args.get(0)), 
                            new FullRectangle(ScriptUtils.getColor(args.get(1)),
                            ScriptUtils.getInteger(args.get(2)), ScriptUtils.getInteger(args.get(3)),
                            ScriptUtils.getInteger(args.get(4)), ScriptUtils.getInteger(args.get(5)))); break;
                case "drawimage":
                    drawImage(args); break;  
                case "drawstring":
                    drawString(args); break;
                case "removegraphic":
                    SnuviScript.screen.graphics.removeGraphicObject(ScriptUtils.getInteger(args.get(0))); break;
                case "buffergraphic":
                    SnuviScript.screen.graphics.loadGraphicObjectIntoBuffer(ScriptUtils.getInteger(args.get(0))); break;
                case "setcolor":
                    SnuviScript.screen.graphics.getBuffer().setColor(ScriptUtils.getColor(args.get(0))); break;
                case "setx":
                    ((StartPoint) SnuviScript.screen.graphics.getBuffer()).setX(ScriptUtils.getInteger(args.get(0))); break;
                case "sety":
                    ((StartPoint) SnuviScript.screen.graphics.getBuffer()).setY(ScriptUtils.getInteger(args.get(0))); break;
                case "setx1":
                    ((Line) SnuviScript.screen.graphics.getBuffer()).setFirstX(ScriptUtils.getInteger(args.get(0))); break;
                case "sety1":
                    ((Line) SnuviScript.screen.graphics.getBuffer()).setFirstY(ScriptUtils.getInteger(args.get(0))); break;
                case "setx2":
                    ((Line) SnuviScript.screen.graphics.getBuffer()).setSecondX(ScriptUtils.getInteger(args.get(0))); break;
                case "sety2":
                    ((Line) SnuviScript.screen.graphics.getBuffer()).setSecondY(ScriptUtils.getInteger(args.get(0))); break; 
                case "setscreensize":
                    SnuviScript.screen.setSize(ScriptUtils.getInteger(args.get(0)), ScriptUtils.getInteger(args.get(1)));
                    SnuviScript.screen.graphics.setSize(SnuviScript.screen.getWidth(), SnuviScript.screen.getHeight() - 22); break;
                case "getscreenx":
                    return SnuviScript.screen.graphics.getWidth();
                case "getscreeny":
                    return SnuviScript.screen.graphics.getHeight();

                // Allgemeines
                case "debug":
                    System.out.println(args.get(0));
                    break;
                case "loadlib":
                    sd.loadLibrary(args.get(0)); break;
                case "lib":
                    SnuviScript.scriptController.startLibrary(qd, args); return false;
                case "reset":
                    sd.resetOverflowProtection(); break;                   
                case "wait":
                    sd.resetOverflowProtection(); return false; 
                case "term":
                    SnuviScript.scriptController.term(sd); return false;
                case "exit":
                    System.exit(0); return false;
                case "loadevent":
                    sd.loadEvent(args.get(0)); break; 
                case "unloadevent":
                    sd.unloadEvent(args.get(0)); break; 
                case "goto":
                    sd.gotoLabel(args.get(0).toString()); break;
                case "sgoto":
                    scheduleGoto(args, sd); break;                    

                // Conditions
                case "if":
                    ifFunction(args, sd); break;   
                case "equal":
                    return ((Comparable) args.get(0)).compareTo(args.get(1)) == 0;
                case "less":
                    return ((Comparable) args.get(0)).compareTo(args.get(1)) == -1;
                case "greater":
                    return ((Comparable) args.get(0)).compareTo(args.get(1)) == 1;
                case "notequal":
                    return ((Comparable) args.get(0)).compareTo(args.get(1)) != 0;
                case "lessequal":
                    return ((Comparable) args.get(0)).compareTo(args.get(1)) <= 0;
                case "greaterequal":
                    return ((Comparable) args.get(0)).compareTo(args.get(1)) >= 0;
                case "invert":
                    return args.stream().allMatch(s -> s.equals(true));
                case "and":
                    return args.stream().allMatch(s -> s.equals(true));
                case "or":
                    return args.stream().anyMatch(s -> s.equals(true));
                    
                // Sonstiges
                case "concatlist":                           
                    return String.join("", ((List) sd.getVar(args.get(0))).subList(ScriptUtils.getInteger(args.get(1)), ScriptUtils.getInteger(args.get(2))));
                case "concatlistspace":                           
                    return String.join(" ", ((List) sd.getVar(args.get(0))).subList(ScriptUtils.getInteger(args.get(1)), ScriptUtils.getInteger(args.get(2))));
                case "split":  
                    sd.setVar(args.get(0), new ArrayList<>(Arrays.asList(args.get(2).toString().split(args.get(1).toString())))); break;
                case "concat":                           
                    return args.stream().map(o -> o.toString()).collect(Collectors.joining());
                case "concatspace":                           
                    return args.stream().map(o -> o.toString()).collect(Collectors.joining(" "));
                case "tolowercase":                           
                    return args.get(0).toString().toLowerCase();
                case "touppercase":                           
                    return args.get(0).toString().toUpperCase();

                case "gettime":                           
                    return System.currentTimeMillis(); 
                case "nextday":        
                    return getNextDay(args, sd);   
                case "waitfor":    
                    return waitFor(args, sd);

                // Is Valid Checker    
                case "isvalidnumber":
                    return generalValidChecker(() -> ScriptUtils.getDouble(args.get(0)));*/

                default:
                    throw new NoSuchMethodException();                      
            }
        }
        catch(Exception ex)
        {
            //ex.printStackTrace();
            printQuestException(sd, ex, function);
            return null;
        }
    }
    
    public static void printQuestException(Script sd, Exception ex, String line)
    {
        sd.resetLoopCounter();
        ScriptUtils.printDebug("Error in");
        ScriptUtils.printDebugList("Script: " + sd.getName());
        ScriptUtils.printDebugList("Zeile: " + line);
        ScriptUtils.printDebugList("Exception: " + ex.getClass().getSimpleName());
        if(ex instanceof IllegalStringException)
        {
            ScriptUtils.printDebugList("Ungültiger Wert: " + ((IllegalStringException) ex).getBadString());
        }
    }
    
    // -----------------------
    // Script Befehle
    // -----------------------
    
    private static void drawImage(List<Object> args) throws IOException 
    {
        BufferedImage image = ImageIO.read(new File(args.get(1).toString()));
        if(args.size() >= 5)
        {
            SnuviScript.screen.graphics.addGraphicObject(ScriptUtils.getInteger(args.get(0)), 
                    new Image(image, ScriptUtils.getInteger(args.get(2)), ScriptUtils.getInteger(args.get(3)),        
                    ScriptUtils.getInteger(args.get(4)), ScriptUtils.getInteger(args.get(5)),
                    ScriptUtils.getInteger(args.get(6)), ScriptUtils.getInteger(args.get(7))));
            return;
        } 
        SnuviScript.screen.graphics.addGraphicObject(ScriptUtils.getInteger(args.get(0)), 
                    new Image(image, ScriptUtils.getInteger(args.get(2)), ScriptUtils.getInteger(args.get(3))));
    }
    
    private static void drawString(List<Object> args) throws IllegalColorException
    {
        SnuviScript.screen.graphics.addGraphicObject(ScriptUtils.getInteger(args.get(0)), 
                new Text(ScriptUtils.getColor(args.get(1)), 
                ScriptUtils.getInteger(args.get(2)), ScriptUtils.getInteger(args.get(3)),        
                args.get(4).toString(), new Font(args.get(5).toString(), ScriptUtils.getInteger(args.get(6)), ScriptUtils.getInteger(args.get(7)))));
    }

    private static void ifFunction(List<Object> args, Script sd)
    {
        if(args.stream().allMatch(s -> s.equals(true)))
        {
            //sd.goDeeper();
        }
    }
                  
    private static void sortList(List<Object> args, Script sd) 
    {     
        //List<Object> list = (List<Object>) sd.getVar(args.get(0));
        //sd.setVar(args.get(0), list.stream().sorted().collect(Collectors.toList()));
    }
                   
    private static void scheduleGoto(List<Object> args, Script qd)
    {
        /*SnuviScheduler.doScheduledTask(() -> 
        {
            if(!sd.isValid())
            {
                return;
            }
            try
            {
                sd.setCodeLine(sd.getGoto(args.get(0)));
                continueCode(qd, null);
            }
            catch(GotoLabelNotFoundException ex)
            {
                ScriptUtilities.printDebugMessage("Das Scheduled-Goto '" + args.get(0) + "' wurde nicht gefunden.");
            }
        }, Integer.parseInt(args.get(1)));*/
    }

    private static String getNextDay(List<Object> args, Script qd)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis(Long.parseLong(args.get(0).toString()));
        cal.set(Calendar.HOUR, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.MILLISECOND, 0);
        cal.add(Calendar.DAY_OF_YEAR, 1);
        return String.valueOf(cal.getTimeInMillis());   
    } 
    
    private static boolean waitFor(List<Object> args, Script qd) throws UnsupportedOperationException
    {           
        /*SnuviScript.scriptController.resetOverflowProtection();
        int i = Integer.parseInt(args.get(0));
        if(i < 1)
        {
            throw new UnsupportedOperationException();
        }
        SnuviScheduler.doScheduledTask(() -> 
        {                   
            if(qd == null || !sd.isValid())
            {
                return;
            }
            sd.setWaiting(false);
            sd.nextCodeLine();
            continueCode(qd, null);
        }, i); 
        sd.setWaiting(true);*/
        return false;
    }

    private static boolean generalValidChecker(Runnable r)
    {
        try
        {
            r.run();
            return true;
        }
        catch(Exception ex)
        {
            return false;
        }
    }   
}
