package me.hammerle.code;

import java.util.HashMap;
import me.hammerle.exceptions.CodeTooLongException;
import me.hammerle.exceptions.GotoLabelNotFoundException;

public class Script 
{
    private final int id;
    private final String name;  

    private final HashMap<String, Object> variables;    
    private final HashMap<String, Integer> gotos;
    
    private final Code[] code;
    private int position;

    private int loopCounter;
    
    public Script(int id, String name, String code)
    {       
        this.id = id;
        this.name = name;
        
        variables = new HashMap<>();
        gotos = new HashMap<>();
        
        this.code = Code.generate(code);
        
        position = 0;

        loopCounter = 0;
    }
    
    // -----------------------------------------------------------------------------------
    // Script-Daten
    // -----------------------------------------------------------------------------------
    
    public int getId() 
    {
        return id;
    }

    public String getName() 
    {
        return name;
    }
    
    // -----------------------------------------------------------------------------------
    // Variablen
    // -----------------------------------------------------------------------------------
    
    public void setVar(String var, Object value)
    {
        variables.put(var, value);   
    }
    
    public Object getVar(String var)
    {
        return variables.get(var);
    }
    
    public void removeVar(String var)
    {
        variables.remove(var);
    }
    
    // -----------------------------------------------------------------------------------
    // Goto
    // -----------------------------------------------------------------------------------
    
    public void gotoLabel(String label) throws GotoLabelNotFoundException, CodeTooLongException
    {
        try
        {
            int i = gotos.get(label);
            loopCounter++;
            if(loopCounter > 50)
            {
                resetLoopCounter();
                throw new CodeTooLongException();
            }
            position = i;
        }
        catch(NullPointerException ex)
        {
            throw new GotoLabelNotFoundException(label);
        }
    }  
    
    public void resetLoopCounter()
    {
        loopCounter = 0;
    }
}
